pub struct Function { /* private fields */ }Expand description
A function wrapper.
Wraps a function by moving it into a closure. Before running, the parameters of the function are constructed from a TypeMap of resources. The results of the function are packed back into the same TypeMap.
Implementations